
import kolossus

# set batch size for how many pairs to run model on at a time
kolossus.set_batch_size(3)


ffasta = 'test.fasta'
fembeds = 'embed.h5'
fpairs = 'pairs.txt'


print(kolossus.kolossus(fpairs, fseqs=ffasta))

print('-' * 40)

print(kolossus.kolossus(fpairs, fembeds=fembeds))




from .utils.seq_to_embedding import extract_embeddings_from_fasta
from .utils.seq_to_embedding import extract_embeddings_from_fasta_individual

import argparse
import sys 
import pathlib
import os 

ESM_TRANSFORMER_LAYERS = {
        "esm2_t6_8M_UR50D": 6,
        "esm2_t12_35M_UR50D": 12,
        "esm2_t30_150M_UR50D": 30,
        "esm2_t33_650M_UR50D": 33,
        "esm2_t36_3B_UR50D": 36,
        "esm2_t48_15B_UR50D": 48
}


def main():
    args = parse_args()
    
    fasta_file = args.i
    model_name = args.model
    device = args.device
    output_file = args.o

    layer = ESM_TRANSFORMER_LAYERS.get(model_name, None)
    if not layer:
        print(f"Error: illegal model name: {model_name}", file=sys.stderr)
        print("Must be one of:", ESM_TRANSFORMER_LAYERS.keys(), file=sys.stderr)
        sys.exit(1)

    if args.individual:
        extract_embeddings_from_fasta_individual(fasta_file, output_file, device, model_name,
                                                 repr_layers=[layer], layer_to_use=layer)
    else:
        extract_embeddings_from_fasta(fasta_file, output_file, device, model_name,
                                      repr_layers=[layer], layer_to_use=layer)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', type=str, required=True, help='name of input fasta file')
    parser.add_argument('--model', type=str, default='esm2_t48_15B_UR50D', help='name of model to extract embeddings')
    parser.add_argument('--device', type=str, default='cpu', help='cpu or gpu device to use')
    parser.add_argument('--individual', action='store_true', default=False, help='flag to indicate whether or not to store individual sequences')
    parser.add_argument('-o', type=str, required=True, help='name of output .h5 file (or output directory if --individual is true)')
    args = parser.parse_args()
    return args 


if __name__ == '__main__':
    main()




from functools import wraps 
from functools import reduce 


class DISCARD_COMPONENTS():

    def __init__(self, len_t, result_is_list, discard_set=None):
        self.discard_set = set() if not discard_set else(set(discard_set))
        self.result_is_list = result_is_list
        self.len_t = len_t

    def __call__(self, func):

        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)

            if self.result_is_list:
                out = [self.remove_comps(x) for x in result]
            else:
                out = self.remove_comps(result)

            return out
        
        return wrapper

    def add_discard_index(self, *args):
        for i in args:
            if i < 0:
                i += self.len_t
            self.discard_set.add(i)

    def remove_comps(self, t):
        return tuple(x for i, x in enumerate(t) if i not in self.discard_set)




import sys 


def main():
    s = "aagctt"

    print(s)
    print('chop by 1:', chop(s, 1))
    print('chop by 2:', chop(s, 2))
    print('chop by 3:', chop(s, 3))
    print('chop by 4:', chop(s, 4))

    subs_valid = [('a', 'b', 0), ('g', 'b', 2)] 
    print("applying valid subs!")
    print(apply_subs(s, subs_valid))

    subs_invalid = [('a', 'b', 0), ('c', 'b', 2)]
    print("applying invalid subs!")
    try:
        print(apply_subs(s, subs_invalid))
        print("This isn't good, should've thrown an assertion error")
        sys.exit(1)
    except AssertionError as e:
        print("Good, raised assertion error!")

    


def chop(s, n):
    i = 0
    out = []
    while i < len(s):
        out.append(s[i:i+n])
        i += n
    return out


def apply_subs(s, subs, zero_based=True):
    """
    Apply substitution mutations to a string. 

    Input:
        - s (str or Bio.Seq.Seq object): sequence to which we apply mutations. 
        - subs [list[tuple]]: each tuple is 3-tuple: (from, to, position)
    """
    # positions to which to apply mutation
    if zero_based:
        submap = {t[2]: (t[0], t[1]) for t in subs}
    else:
        submap = {t[2]-1: (t[0], t[1]) for t in subs}

    # output variable 
    out = []

    # remember to add a try-catch for invalid subs 
    for i, c in enumerate(s):
        a, b = submap.get(i, (c, c))
        assert s[i] == a
        out.append(b)

    return ''.join(out)


if __name__ == '__main__': 
    main()



#### Utilities for batching a computation. ####


# assumes that `func` is a function that operates on a list
def batch_comp_on_list(block_size=10000):

    def decorator(func):

        def inner(data, *args, **kwargs):
            nobs = len(data)
            lo = 0
            niter = (nobs // block_size) + (0 if nobs % block_size == 0 else 1)
            results = [None for _ in range(niter)]
            ct = 0
            while lo < nobs:
                hi = min(lo + block_size, nobs)
                print(f"computing results for observations {lo} through {hi-1}...", end=' ')
                results[ct] = func(data[lo:hi], *args, **kwargs)
                print("done!")
                lo = hi
                ct += 1
            return results

        return inner

    return decorator 





from .seqops import chop


def write_fasta(seqs, outfile, n=80):
    """
    Write sequences to fasta file. Note that we assume the order of the sequences
    is unimportant if the input passed in is a dictionary. 

    Input: 
        seqs (dict[str] -> str) OR (list[(str, str)]): a dictionary or a list of tuples (fasta header, sequence)
        outfile (str): path for output file
    """
    if isinstance(seqs, dict):
        seqs = seqs.items()
    with open(outfile, 'wt') as fout:
        for header, seq in seqs:
            print(f'>{header}', file=fout)
            print('\n'.join(chop(seq, n)), file=fout)



import time 
from functools import wraps


def main():
    @delay(0.3, 0.0)
    def foo(tsleep):
        time.sleep(tsleep)
        print("hello")

    for _ in range(int(3)):
        foo(0.2)


def delay(limit=1.0, time_delay=1.0):

    def decorator(func):
        # initializing to limit will ensure no delay on the first call
        state = {'prev_call': limit}

        @wraps(func)
        def wrapper(*args, **kwargs):

            # get current time 
            now = time.time()
            prev_call = state['prev_call']
            if abs(now - prev_call) < limit:
                time.sleep(time_delay)

            result = func(*args, **kwargs)

            state['prev_call'] = time.time()

            return result

        return wrapper

    return decorator


def fn_timer(func):
    @wraps(func)
    def wrapper(*args, **kwargs):

        # run and time function
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        elapsed_time = end_time - start_time

        print(f"{func.__name__} took {elapsed_time:.3f} seconds to run.")
        return result
    return wrapper


if __name__ == '__main__': 
    main()




from Bio import SeqIO


def parse_fasta(fname, to_dict=False):
    if to_dict:
        return {r.id: str(r.seq) for r in SeqIO.parse(fname, 'fasta')}
    else:
        return [(r.id, str(r.seq)) for r in SeqIO.parse(fname, 'fasta')]





import torch
from .Network import Network

# note: very hacky, change to host on some website in the future 
import os
import sys 


MODEL_FILE = 'saved_model_cs_ws_ft_humansty.epoch_59.pth'
FILE_PATH = os.path.join(os.path.split(__file__)[0], MODEL_FILE)
FILE_HUGGINGFACE_URL = 'https://huggingface.co/aparekh2/kolossus/resolve/main/saved_model_cs_ws_ft_humansty.epoch_59.pth'


if not os.path.isfile(os.path.join(os.path.split(__file__)[0], MODEL_FILE)):
    print(f"Downloading model from {FILE_HUGGINGFACE_URL} to {FILE_PATH}.")
    torch.hub.download_url_to_file(FILE_HUGGINGFACE_URL, FILE_PATH)


def load_model():
    path = os.path.split(__file__)[0]
    model_weights = os.path.join(path, MODEL_FILE)
    model = Network()
    model.load_state_dict(torch.load(model_weights))
    return model 
